;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LED Toy 24, User Interface
;
; Written 2005 by Werner Almesberger
; Copyright 2005 by Werner Almesberger
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


main:

	bsf	STATUS,RP0
	bcf	RX 		; make RX an output (LED)
	bsf	TX		; make TX an input (button)
	bcf	STATUS,RP0

	call	lightsout	; turn off all lights

	bsf	INTCON,PEIE	; enable peripheral interrupts

; Initialize configuration variables

	movlw   CFG_NAP
	movwf   nap
	call    updnap

	movlw   CFG_GAP
	movwf   gap

	movlw   CFG_SHOWT
	movwf   showt
	call    updshowt

	call	select0		; select the first object

; Run through all LEDs

	movlw	NUM_LEDS	; all LEDs
	movwf	tmp

mlloop:
	movf	tmp,W		; LED is NUM_LEDS-tmp
	sublw	NUM_LEDS
	call	setwth

	movlw	W_RUNNING	; hold LED
	call	waitw

	decfsz	tmp,F		; repeat
	goto	mlloop

	; fall through to zzz

zzz:
	bcf	INTCON,GIE	; no interrupts from here on

if (HW_SINK)
	bsf	STATUS,RP0
	movlw	0xff
	movwf	TRISA
	movwf	TRISB
	bcf	STATUS,RP0
	call	lightsout	; turn off all lights
else
	call	lightsout	; turn off all lights
endif

	bcf	T2CON,TMR2ON	; disable both timers
	bcf	PIE1,TMR2IF
	bcf	INTCON,TMR0IE

zzz_again:
	bsf	STATUS,RP0	; enable pull-up
	bcf	OPTION_REG,NOT_RBPU
	bcf	STATUS,RP0

	movf	PORTB,W		; wake on keypress
	bcf	INTCON,RBIF
	bsf	INTCON,RBIE

	sleep			; take a nap

	call	debounce	; make sure we don't catch release noise

	call	button		; button pressed ?
	btfss	STATUS,C
	goto	zzz_again	; no -> sleep some more

if (HW_SINK)
	bsf	STATUS,RP0
	clrf	TRISA
	movlw	1 << 2
	movwf	TRISB
	bcf	STATUS,RP0
endif

	bcf	INTCON,RBIE	; disable button change interrupts

	bsf	STATUS,RP0
	bsf	RX		; make RX an input
	bcf	TX		; make TX an output
	bcf	STATUS,RP0

	call	monvec		; maybe enter the monitor

	bsf	STATUS,RP0	; clean up after the monitor
	bcf	RX		; make RX an output (LED)
	bsf	TX		; make TX an input (button)
	bcf	STATUS,RP0

	bcf	CURRBUT		; clear button flag (= pressed)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; UI state machine
;

;
; Any state with button change or timeout
;

uistate	macro	wait_action,timeout,keystate,tostate
	local	uiloop

	call	timeout
uiloop:
	call	wait_action
	call	buttonq		; button change ?
	btfsc	STATUS,C	; yes -> debounce and go
	goto	keystate
	call	toq		; timeout ?
	btfss	STATUS,Z	; no -> keep on looping
	goto	uiloop
	goto	tostate		; timeout state
	endm

;
; State showing the image, without a visible counter
;

uiimgop	macro	action,outstate
	local	uiin,uidown,uichange,uiout

	uistate	act_image,t_idle,uiin,ui_off			; down
uiin:
	uistate	act_image,t_config,uidown,ui_off		; up
uidown:
	uistate	act_image,t_long,uichange,uiout			; down
uichange:
	call	action
	goto	uiin
uiout:
	uistate	act_clear,t_idle,outstate,ui_off		; down
	endm

;
; State showing the image, with a visible counter
;

uiimgct	macro	action,var,outstate
	local	uiin,uiimg,uidown,uichange,uichnou,uiout

	uistate	act_image,t_idle,uiin,ui_off			; down
uiin:
	movf	var,W		; show the counter
	call	setwth
	uistate	act_noop,t_current,uidown,uiimg			; up
uiimg:
	uistate	act_image,t_config,uidown,ui_off		; up
uidown:
	uistate	act_image,t_long,uichange,uiout			; down
uichange:
	movlw	-1		; decrement counter
	addwf	var,F
	btfsc	STATUS,C	; no underrun
	goto	uichnou
	movlw	NUM_LEDS-1	; wrap counter
	movwf	var
uichnou:
	call	action
	goto	uiin
uiout:
	uistate	act_clear,t_idle,outstate,ui_off		; down
	endm


;
; State showing just a counter
;

uicnt	macro	action,var,outstate
	local	uiin,uidown,uichange,uichnou,uiout

	uistate	act_clear,t_idle,uiin,ui_off			; down
uiin:
	movf	var,W		; show the counter
	call	setwth
	uistate	act_noop,t_config,uidown,ui_off			; up
uidown:
	uistate	act_noop,t_long,uichange,uiout			; down
uichange:
	movlw	-1		; decrement counter
	addwf	var,F
	btfsc	STATUS,C	; no underrun
	goto	uichnou
	movlw	NUM_LEDS-1	; wrap counter
	movwf	var
uichnou:
	call	action
	goto	uiin
uiout:
	uistate	act_clear,t_idle,outstate,ui_off		; down
	endm


;
; The menu
;

	uistate	act_led_run,t_long,ui_run_ind,adc_test		; down
ui_run_ind:
	uistate	act_led_run,t_idle,ui_run_sel,ui_off		; up
ui_run_sel:
	uistate	act_led_run,t_long,ui_select_ind,ui_run		; down
ui_select_ind:
	uistate	act_led_select,t_idle,ui_select_sel,ui_off	; up
ui_select_sel:
	uistate	act_led_select,t_long,ui_rotate_ind,ui_select	; down
ui_rotate_ind:
	uistate	act_led_rotate,t_idle,ui_rotate_sel,ui_off	; up
ui_rotate_sel:
	uistate	act_led_rotate,t_long,ui_freq_ind,ui_rotate	; down
ui_freq_ind:
	uistate	act_led_freq,t_idle,ui_freq_sel,ui_off		; up
ui_freq_sel:
	uistate	act_led_freq,t_long,ui_space_ind,ui_freq	; down
ui_space_ind:
	uistate	act_led_space,t_idle,ui_space_sel,ui_off	; up
ui_space_sel:
	uistate	act_led_space,t_long,ui_to_ind,ui_space		; down
ui_to_ind:
	uistate	act_led_to,t_idle,ui_to_sel,ui_off		; up
ui_to_sel:
	uistate	act_led_to,t_long,ui_run_ind,ui_to		; down


;
; Run
;

ui_run:
	movf	seqph,F		; run a sequence ?
	btfss	STATUS,Z
	goto	ui_run_single	; no
	movf	showtl,W	; to1[hl] = showt[hl]
	movwf	to1l
	movf	showth,W
	movwf	to1h
seq0:	movlw	2		; start sequence
	movwf	seqind
seql:
	movf	showtl,W	; set the overall timeout
	movwf	to1l
	movf	showth,W
	movwf	to1h

	movlw	seqpl		; load sequence pointer
	call	eesetp
	call	eebytel		; skip the title object

	
ui_run_single:
	uistate	act_image,t_config,ui_run_wait,ui_off		; down
ui_run_wait:
	uistate	act_image,t_timeout,ui_run_but,ui_off		; up
ui_run_but:
	uistate	act_image,t_long,ui_run_wait,ui_off		; down


;
; Select
;

ui_select:
	uiimgop	act_select,ui_select_ind

select0:
	clrf	objind		; clear the object index
	btfss	STATUS,Z
act_select:
	incf	objind,F	; get the next object pointer
	call	objindee
	call	eeptr
	btfsc	STATUS,Z	; wrap back to the first image ?
	goto	select0
	clrf	wideph		; clear wide image pointer
	clrf	seqph		; clear the sequence base pointer
selobj:
	call	eebytel		; read the type
	movwf	widex		; (store width in case it's a wide image)
	call	eebyteh
	btfsc	STATUS,Z	; regular image ?
	goto	copyimg		; -> copy the image to RAM
	xorlw	OBJ_SEQ		; sequence ?
	btfss	STATUS,Z
	goto	selnoseq
	movlw	seqpl		; save the sequence pointer
	call	eegetp
	call	eeptr		; -> load the title object
	goto	selobj
selnoseq:
	movlw	widepl		; store wide image pointer
	goto	eegetp

;
; Rotate
;

ui_rotate:
	uiimgop	act_rotate,ui_rotate_ind

act_rotate:
	goto	imgrot


;
; Frequency
;

ui_freq:
	uiimgct	act_freq,nap,ui_freq_ind

act_freq:
	movf	nap,W
	goto	updnap

;
; Space
;

ui_space:
	uiimgct	act_noop,gap,ui_space_ind


;
; Timeout
;

ui_to:
	uicnt	act_to,showt,ui_to_ind

act_to:
	movf	showt,W
	goto	updshowt


ui_off:
	goto	zzz


;
; LED actions
;

act_led_run:
	movlw	L_RUN
	goto	togwth

act_led_select:
	movlw	L_SELECT
	goto	togwth

act_led_rotate:
	movlw	L_ROTATE
	goto	togwth

act_led_freq:
	movlw	L_FREQ
	goto	togwth

act_led_space:
	movlw	L_GAP
	goto	togwth

act_led_to:
	movlw	L_TO
	goto	togwth


;
; Display actions
;

act_clear:			; turn off all LEDs (and phase transistors)
lightsout:
if (HW_SINK)
	movlw	0xff
	movwf	PORTA
	movwf	PORTB
else
	clrf	PORTA
	clrf	PORTB
endif
	return

act_image:
	movf	wideph,F	; do we have a wide image ?
	btfsc	STATUS,Z	; no -> display regular image
	goto	doframe
	movlw	widepl		; set the image pointer
	call	eesetp
	movf	widex,W		; get the image size
	goto	doframeee	; display it

act_noop:
	return


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Timer handling
;


;
; "Long" timeout
;

t_long:
	movlw	T_LONG
	goto	arm0


;
; "Idle" timeout                 
;

t_idle:
	movlw	T_IDLE
	goto	arm0


;
; "Show current" timeout
;

t_current:
	movlw	T_CURRENT
	goto	arm0


;
; "Configuration idle" timeout
;

t_config:
	movlw	low T_CONFIG
	movwf	to1l
	movlw	high T_CONFIG
	movwf	to1h
	goto	arm


;
; "Show" timeout
;

t_timeout:
	movf	showtl,W	; copy the timeout value
	movwf	to1l
	movf	showth,W
	movwf	to1h
	goto	arm


;
; Arm the timer
;

arm0:
	movwf	to1l		; lower timeout value in W
	clrf	to1h		; higher value is constant (0)
arm:
	clrf	to2l		; no secondary timeout
	clrf	to2h
	bsf	STATUS,RP0	; run ar OSC/4
	bcf	OPTION_REG,T0CS
	bcf	OPTION_REG,PSA	; the prescaler is ours
	; prescaler defaults to 256 on reset
	bcf	STATUS,RP0

	clrf	TMR0		; 15.259 Hz (pretend it's 16 Hz)

	bcf	INTCON,TMR0IF	; clear any pending interrupts

	return


;
; Check if we have a timeout. Return with Z set if we do. whichto:0 is set if
; the timeout came from the second timer.
;

toq:
	btfss	INTCON,TMR0IF	; timer expired ?
	goto	toqno2
	bcf	bouncing,0	; button is now debounced
	clrf	whichto		; clear whichto:0
	bcf	INTCON,TMR0IF	; re-arm timer
	movf	to1l,W		; never time out (1st) ?
	iorwf	to1h,W
	btfsc	STATUS,Z
	goto	toqno1
	movlw	-1		; decrement ~16 Hz counter
	addwf	to1l,F
	btfsc	STATUS,C	; underflow ?
	goto	toqnu1
	addwf	to1h,F		; decrement high byte
	btfsc	STATUS,C	; underflow ?
	goto	toqno1
	bsf	STATUS,Z	; timeout
	return
toqno1:
	bsf	whichto,0	; now, only the 2nd timer can cause a timeout
	movf	to2l,W		; never time out (2nd) ?
	iorwf	to2h,W
	btfsc	STATUS,Z
	goto	toqno2
	movlw	-1		; decrement ~16 Hz counter
	addwf	to2l,F
	btfsc	STATUS,C	; underflow ?
	goto	toqnu2
	addwf	to2h,F		; decrement high byte
	btfsc	STATUS,C	; underflow ?
	goto	toqno2
	bsf	STATUS,Z	; timeout
	return
toqno2:
	bcf	STATUS,Z	; no timeout
	return
toqnu1:
	btfsc	STATUS,Z	; at 0 ?  (if not, return with Z clear)
	movf	to1h,F		; to1h determines Z
	return
toqnu2:
	btfsc	STATUS,Z	; at 0 ?  (if not, return with Z clear)
	movf	to2h,F		; to2h determines Z
	return


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Button handling
;


;
; Check if the button has changed. Set C if it did.
;

buttonq:
	bcf	STATUS,C	; default is no change
	btfsc	bouncing,0	; is the button still bouncing ?
	return
	btfss	CURRBUT		; probe depends on current button
	goto	butqz
	call	button		; test for change
	btfss	STATUS,C
	return
	bcf	CURRBUT		; update "last button"
	goto	butqch		; process change

butqz:
	call	button		; test for change
	btfss	STATUS,C
	goto	butqdn
	bcf	STATUS,C	; no change
	return

butqdn:
	bsf	CURRBUT		; update "last button"
butqch:
	bsf	STATUS,C	; indicate change
	bsf	bouncing,0	; start debouncing
	return


; Debounce button

debounce:
	movlw	W_DEBOUNCE	; for the button to settle
	goto	waitw


;
; Query the button. Set C if it is pressed.
;

button:
        bsf	STATUS,RP0	; enable pull-ups
	bcf	OPTION_REG,NOT_RBPU
        bcf	STATUS,RP0
	
	bcf	STATUS,C	; default is not pressed
	btfss	BUTTON
	bsf	STATUS,C	; set C if pressed

        bsf	STATUS,RP0	; disable pull-ups
	bsf	OPTION_REG,NOT_RBPU
        bcf	STATUS,RP0

	return


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Show one frame from RAM
;

qout	macro
if (HW_SINK)
	movlw	0xff
	movwf	PORTA
	movwf	PORTB
else
	nop
	clrf	PORTA
	clrf	PORTB
endif
	endm

doframe:
	bsf	STATUS,IRP	; access image memory (0x110-0x16f)
	movlw	0x10
	movwf	FSR

	movlw	NUM_LEDS
	movwf	rows
dorow:				; copy four bytes to scratch registers
	movf	INDF,W
	incf	FSR,F
	movwf	a0
	movf	INDF,W
	incf	FSR,F
	movwf	b0
	movf	INDF,W
	incf	FSR,F
	movwf	a1
	movf	INDF,W
	incf	FSR,F
	movwf	b1

	movf	napiter,W	; n*20 us
	movwf	tmp

cycle:
	qout			; 3

	movf	a0,W		; 10
	movwf	PORTA
	movf	b0,W
	movwf	PORTB
	nop
	nop
	nop
;	nop
;	nop
;	nop

	qout			; 3

	movf	b1,W		; 10
	movwf	PORTB
	movf	a1,W
	movwf	PORTA
;	nop
;	nop
;	nop
	decfsz	tmp,F
	goto	cycle

	call	lightsout	; clear all LEDs

	decfsz	rows,F
	goto	dorow

	; fall through to showgap

showgap:			; show the inter-frame gap

	movf	gap,W		; get the gap size
	btfsc	STATUS,Z	; no gap ?
	return

	movwf	rows		; use "rows" counter

gappx:
	movf	napiter,W	; 2*n*20 us
	movwf	tmp
	addwf	tmp,F

gapcy:
	nop			; 7
	nop
	nop
	nop
	nop
	nop
	nop
	decfsz	tmp,F
	goto	gapcy

	decfsz	rows,F		; 3
	goto	gappx

	return			; frame shown


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Show one frame from Flash. Starting position is EEADR{,H}. Width is in W.
;

doframeee:
	movwf	rows
dorowee:			; copy four bytes to scratch registers
	call	eebytel
	movwf	a0
	call	eebyteh
	movwf	b0
	btfsc	a0,5		; RA5 -> RB6
	bsf	b0,6
	btfsc	b0,5		; RB5 -> RB7
	bsf	b0,7
	call	eebytel
	movwf	a1
	call	eebyteh
	movwf	b1
	btfsc	a1,5		; RA5 -> RB6
	bsf	b1,6
	btfsc	b1,5		; RB5 -> RB7
	bsf	b1,7

	movf	napiter,W	; n*20 us
	movwf	tmp

cycleee:
	qout			; 2

	movf	a0,W		; 10
	movwf	PORTA
	movf	b0,W
	movwf	PORTB
	nop
	nop
	nop
	nop
	nop
	nop

	qout			; 2

	movf	b1,W		; 10
	movwf	PORTB
	movf	a1,W
	movwf	PORTA
	nop
	nop
	nop
	decfsz	tmp,F
	goto	cycleee

	call	lightsout	; clear all LEDs

	decfsz	rows,F
	goto	dorowee

	goto	showgap		; show the inter-frame gap and return


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Copy image from current EEPROM position to RAM
;

copyimg:
	bsf	STATUS,IRP	; access image memory (0x110-0x16f)
	movlw	0x10
	movwf	FSR

	movlw	NUM_LEDS*2	; 24 rows, 2 words each
	movwf	tmp

copybyte:
	call	eebytel		; fetch the lower byte of the word (port A)

	movwf	INDF		; store in RAM

	call	eebyteh		; fetch the higher byte (port B)

	btfsc	INDF,5		; RA5 -> RB6
	iorlw	1 << 6

	incf	FSR,F		; store the port B byte
	movwf	INDF

	btfsc	INDF,5		; RB5 -> RB7
	bsf	INDF,7
	bcf	INDF,5

	incf	FSR,F		; next byte

	decfsz	tmp,F		; next word
	goto	copybyte

	return			; done


;
; W = low(*eeprom++); Z reflects value of W.
;

eebytel:
	bsf	STATUS,RP1
	bsf	STATUS,RP0
	clrf	EECON1
	bsf	EECON1,EEPGD
	bsf	EECON1,RD
	nop
	nop
	bcf	STATUS,RP0
	incfsz	EEADR,F
	goto	eenowrap
	incf	EEADRH,F
eenowrap:
	movf	EEDATA,W
	bcf	STATUS,RP1
	return

;
; W = high(previous); Z reflects value of W.
;

eebyteh:
	bsf	STATUS,RP1
	movf	EEDATH,W
	bcf	STATUS,RP1
	return

;
; Set Flash pointer from register pair pointed to by W
;

eesetp:
	bcf	STATUS,IRP	; access regular memory
	movwf	FSR
	bsf	STATUS,RP1	; EEADR = *FSR
	movf	INDF,W
	movwf	EEADR
	incf	FSR,F		; FSR++
	movf	INDF,W		; EEADRH = *FSR
	movwf	EEADRH
	bcf	STATUS,RP1
	return

;
; Store Flash pointer in register pair pointed to by W
;

eegetp:
	bcf	STATUS,IRP	; access regular memory
	movwf	FSR
	bsf	STATUS,RP1	; EEADR = *FSR
	movf	EEADR,W
	movwf	INDF
	incf	FSR,F		; FSR++
	movf	EEADRH,W	; EEADRH = *FSR
	movwf	INDF
	bcf	STATUS,RP1
	return

;
; EEADR{,H} = *(EEADR{,H})++, returns with Z set if EEADRH = 0
;

eeptr:
	call	eebytel
	bsf	STATUS,RP1
	movwf	EEADR
	movf	EEDATH,W
	movwf	EEADRH
	bcf	STATUS,RP1
	return


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Object pointer handling
;

;
; Set the EEPROM pointer from the object index
;

objindee:
	movf	objind,W	; add index to start address
	addlw	low datastart
	bsf	STATUS,RP1
	movwf	EEADR
	movlw	high datastart
	btfsc	STATUS,C	; carry from low to high
	addlw	1
	movwf	EEADRH
	bcf	STATUS,RP1
	return			; done


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Sequence pointer handling
;

;
; Reset the sequence pointer
;

resetseq:
	movlw	low (datastart+1)
	movwf	seqpl
	movlw	high (datastart+1)
	movwf	seqph

;
; Check if the sequence pointer is at the SEQ_END entry. Returns with Z clear
; if so.
;
seqmore:
	call	loadseqp	; load the sequence pointer

	call	eebytel		; clear Z if at SEQ_END
	call	eebyteh
	andlw	SEQ_END
	return

;
; Move sequence pointer past one sequence
;

skipseq:
	call	loadseqp	; load the sequence pointer
	call	eebytel		; skip the title image

skipstep:
	call	eebytel		; get the first word (image number and rotation)
	call	eebyteh
	andlw	SEQ_OFF | SEQ_LOOP ; any end marker ?
	btfsc	STATUS,Z
	goto	seqend

	call	eebytel		; skip timeout/gap word
	goto	skipstep

seqend:
	call	eebytel		; skip the timeout word

	bsf	STATUS,RP1	; store the sequence pointer
	movf	EEADR,W
	bcf	STATUS,RP1
	movwf	seqpl
	bsf	STATUS,RP1
	movf	EEADRH,W
	bcf	STATUS,RP1
	movwf	seqph

	return

;
; Load the sequence pointer
;

loadseqp:
	movf	seqpl,W
	bsf	STATUS,RP1
	movwf	EEADR
	bcf	STATUS,RP1
	movf	seqph,W
	bsf	STATUS,RP1
	movwf	EEADRH
	bcf	STATUS,RP1

	return


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Light LED (W) or (W+1), depending on toggle
;

togwth:
	btfsc	toggle,0	; toggle is 0
	addlw	1
	comf	toggle,F	; invert toggle

	; fall through to "setwth"


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 
; Set the W-th LED (first LED is 0)
;

setwth:
	movwf	tmp2		; save W
	call	lightsout	; turn off all LEDs (and phase transistors)
	btfsc	tmp2,0		; set phase transistor
	goto	setwp1
if (HW_SINK)
	bcf	PHASE1
else
	bsf	PHASE1
endif
	goto	setwph
setwp1:
if (HW_SINK)
	bcf	PHASE0
else
	bsf	PHASE0
endif
	bcf	tmp2,0		; clear bit0
setwph:
	call	ledwtab		; get port A setting
if (HW_SINK)
	xorlw	0xff
	andwf	PORTA,F
else
	iorwf	PORTA,F
endif
	incf	tmp2,F		; get port B setting
	call	ledwtab
if (HW_SINK)
	xorlw	0xff
	andwf	PORTB,F
else
	iorwf	PORTB,F
endif
	return			; done

ledwtab:
	movlw	high ledwtab	; prepare the jump
	movwf	PCLATH
	movf	tmp2,W		; do lookup
	addwf	PCL,F
	retlw	0		; RB1
	retlw	1 << 1

if (HW_SWAP_RB2_RB3)
	retlw	0		; RB2
	retlw	1 << 2
	retlw	0		; RB3
	retlw	1 << 3
else
	retlw	0		; RB3
	retlw	1 << 3
	retlw	0		; RB2
	retlw	1 << 2
endif

if (HW_SWAP_RB4_RB6)
	retlw	0		; RB4
	retlw	1 << 4
	retlw	0		; RB6
	retlw	1 << 6
else
	retlw	0		; RB6
	retlw	1 << 6
	retlw	0		; RB4
	retlw	1 << 4
endif

	retlw	0		; RB7
	retlw	1 << 7
	retlw	1 << 6		; RA6
	retlw	0
	retlw	1 << 7		; RA7
	retlw	0
	retlw	1 << 0		; RA0
	retlw	0
	retlw	1 << 1		; RA1
	retlw	0
	retlw	1 << 2		; RA2
	retlw	0
	retlw	1 << 3		; RA3
	retlw	0

; Fail if the table above wraps around a page boundary

	org	low $+(256*high ledwtab)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Image rotation
;

;
; If the image looks like this
;
;     x-->
;     0          n         2n
; y 0 +----------+----------+
; |   |          |          |
; v   |          |          |
;     |    Q1    |    Q2    |
;     |          |          |
;     |          |          |
;   n +----------+----------+
;     |          |          |
;     |          |          |
;     |    Q4    |    Q3    |
;     |          |          |
;     |          |          |
;  2n +----------+----------+
;
; we go through all pixels in quadrant 1, rotate them to quandrant 2. Likewise,
; we proceed by rotating the replaced pixels in quadrant 2 to quadrant 3, etc.,
; until we're back and quadrant 1.
;

imgrot:
	movlw	(IMG_EDGE-1)/2	; cycle through half of the rows
	movwf	ry
irylp:
	movlw	(IMG_EDGE-1)/2	; cycle through half of the columns
	movwf	rx
irxlp:

	call	swapbit		; bit (x,y)
	call	swapxy		; (x,y) -> (y,x)

	; We calculate (foo_max-1)-foo as follows:
	;
	; foo_max-1-foo = foo_max-1+(-foo) = foo_max-1+(~foo+1) = foo_max+~foo

	comf	rx,F		; bit (y_max-y,x)
	movlw	IMG_EDGE
	addwf	rx,F
	call	swapbit
	call	swapxy		; (y_max-y,x) -> (x,y_max-y)

	comf	rx,F		; bit (x_max-x,y_max-y)
	movlw	IMG_EDGE
	addwf	rx,F
	call	swapbit
	call	swapxy		; (x_max-x,y_max-y) -> (y_max-y,x_max-x)

	comf	rx,F		; bit (y,x_max-x)
	movlw	IMG_EDGE
	addwf	rx,F
	call	swapbit
	call	swapxy		; (y,x_max-x) -> (x_max-x,y)

	comf	rx,F		; bit (x,y) again
	movlw	IMG_EDGE
	addwf	rx,F
	call	swapbit

	movlw	-1		; next column
	addwf	rx,F
	btfsc	STATUS,C
	goto	irxlp

	movlw	-1		; next row
	addwf	ry,F
	btfsc	STATUS,C
	goto	irylp

	return			; done

; Swap rx and ry

swapxy:
	movf	rx,W
	movwf	rtmp
	movf	ry,W
	movwf	rx
	movf	rtmp,W
	movwf	ry
	return

; Swap bit at (rx,ry) with bit 0 of rbit

swapbit:
	call	selpix		; select the pixel
	btfsc	rbit,0		; set ?
	goto	swbit1

	andwf	INDF,W		; W &= *FSR
	btfsc	STATUS,Z	; W == 0 -> pixel was already 0
	return

	xorwf	INDF,F		; *FSR ^= W
	bsf	rbit,0		; rbit = 1

	return			; done

swbit1:
	movwf	rtmp		; W = (W ^ *FSR) & W
	xorwf	INDF,W
	andwf	rtmp,W
	btfsc	STATUS,Z	; W == 0 -> pixel was already 1
	return

	xorwf	INDF,F		; *FSR ^= W
	bcf	rbit,0		; rbit = 0

	return			; done


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Select the pixel in (rx,ry). INDF points to the byte containing the pixel,
; and W contains a bit mask with a "1" at the position of the pixel.
;

selpix:
	bsf	STATUS,IRP	; FSR offset is 0x100

	movf	rx,W		; FSR = 0x10+4*rx;
	addwf	rx,W
	addlw	8
	movwf	FSR
	addwf	FSR,F

	movf	ry,W		; rtmp = ry:
	movwf	rtmp

	bcf	STATUS,C	; select the phase
	rrf	rtmp,F
	movlw	2		; prepare for phase selection
	btfsc	STATUS,C
	addwf	FSR,F		; if (ry & 1) FSR += 2;

	movlw	-6		; select the port
	addwf	rtmp,W		; < 6 is port B
	btfss	STATUS,C
	goto	spxpb

	btfsc	STATUS,Z	; 0
	retlw	1 << 6		; RA6

	addlw	-1		; 1
	btfsc	STATUS,Z
	retlw	1 << 7		; RA7

	addlw	-1		; 2
	btfsc	STATUS,Z
	retlw	1 << 0		; RA0

	addlw	-1		; 3
	btfsc	STATUS,Z
	retlw	1 << 1		; RA1

	addlw	-1		; 4
	btfsc	STATUS,Z
	retlw	1 << 2		; RA2

	retlw	1 << 3		; 5, RA3

spxpb:
	incf	FSR,F		; FSR++;

	addlw	1		; 5
	btfsc	STATUS,Z
	retlw	1 << 7		; RB7

if (HW_SWAP_RB4_RB6)

	addlw	1		; 4
	btfsc	STATUS,Z
	retlw	1 << 6		; RB6

	addlw	1		; 3
	btfsc	STATUS,Z
	retlw	1 << 4		; RB4

else

	addlw	1		; 4
	btfsc	STATUS,Z
	retlw	1 << 4		; RB4

	addlw	1		; 3
	btfsc	STATUS,Z
	retlw	1 << 6		; RB6

endif

if (HW_SWAP_RB2_RB3)

	addlw	1		; 2
	btfsc	STATUS,Z
	retlw	1 << 3		; RB3

	addlw	1		; 1
	btfsc	STATUS,Z
	retlw	1 << 2		; RB2

else

	addlw	1		; 2
	btfsc	STATUS,Z
	retlw	1 << 2		; RB2

	addlw	1		; 1
	btfsc	STATUS,Z
	retlw	1 << 3		; RB3

endif

	retlw	1 << 1		; 0, RB1


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Convert nap value in W to number of iterations
;
; We use roughly 1.2215^W
;

updnap:
	movwf	tmp		; save W while we set PCLATH
	movlw	high cntab
	movwf	PCLATH
	movf	tmp,W
	call	convnap		; convert nap index to number of iterations
	movwf	napiter
	return

convnap:
	addwf	PCL,F
cntab:
	retlw	1		; 0, 20 us
	retlw	2		; 1, 40 us
	retlw	3		; 2, 60 us
	retlw	4		; 3, 80 us
	retlw	5		; 4, 100 us
	retlw	6		; 5, 120 us

	retlw	7		; 6, 140 us
	retlw	8		; 7, 160 us
	retlw	9		; 8, 180 us
	retlw	10		; 9, 200 us
	retlw	11		; 10, 220 us
	retlw	12		; 11, 240 us

	retlw	13		; 12, 260 us
	retlw	14		; 13, 280 us
	retlw	16		; 14, 320 us	; now we're exponential
	retlw	20		; 15, 400 us
	retlw	24		; 16, 480 us
	retlw	30		; 17, 600 us

	retlw	37		; 18, 740 us
	retlw	45		; 19, 900 us
	retlw	55		; 20, 1100 us
	retlw	67		; 21, 1340 us
	retlw	82		; 22, 1640 us
	retlw	100		; 23, 2000 us

; Fail if the table above wraps around a page boundary

	org	low $+(256*high convnap)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Convert the "show" timeout in W from minutes to 1/16 seconds
; 
; We multiply with 60*16 = 2^6*3*5

updshowt:
	movwf	showtl		; store W in showtl/h
	clrf	showth
	bcf	STATUS,C	; no junk, please
	rlf	showtl,F	; *(4...
	rlf	showtl,F
	addwf	showtl,F	; ...+1) (doesn't set C)
	movf	showtl,W	; copy showtl to W
	rlf	showtl,F	; *(2...
	rlf	showth,F
	addwf	showtl,F	; ...+1)
	btfsc	STATUS,C
	incf	showth,F
	bcf	STATUS,C	; no junk, please

	movlw	6		; *2^6
	movwf	tmp
cvstlp:
	rlf	showtl,F	; *2
	rlf	showth,F
	decfsz	tmp,F
	goto	cvstlp

	return			; done


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; ADC test
;

;
; The analog signal should be in the range 0-1.7V with Vcc = 3V, so we use a
; bit more than half the ADC range. The output routine below should yield a
; value in the range [0,18]. We're extra-careful and limit it to NUM_LEDS-1,
; just in case.
;


adc_test:
	call	button		; button pressed ?
	btfss	STATUS,C
	goto	ui_off		; no -> power down

	; prepare AN2 for analog input

	bsf	STATUS,RP0
	bsf	TRISA,2		; turn off digital output
	movlw	1 << 2		; enable analog input
	movwf	ANSEL
	clrf	ADCON1		; left justified
	bcf	STATUS,RP0

	movlw	(1 << ADCS0) | (2 << CHS0) | (1 << ADON)
				; Fosc/8, select AN2, ADC on
	movwf	ADCON0

	; power up the sensor

if (HW_SINK)
	call	lightsout
else
	clrf	PORTA
	clrf	PORTB
	bsf	PORTA,3
endif

	movlw	5		; wait 501 us
	call	waitw

	bsf	ADCON0,GO	; start the conversion
adclp:
	btfsc	ADCON0,NOT_DONE	; wait until done
	goto 	adclp

	bcf	PORTA,3		; power down the sensor
	bsf	STATUS,RP0	; enable digital output again
	bcf	TRISA,2
	clrf	ANSEL		; disable analog input
	bcf	STATUS,RP0
	clrf	ADCON0		; turn off the ADC

	rrf	ADRESH,F	; get upper 5 bits
	rrf	ADRESH,F
	rrf	ADRESH,F
	movlw	0x1f
	andwf	ADRESH,F
	movlw	-(NUM_LEDS-1)	; cut off range to [0,NUM_LEDS-1]
	addwf	ADRESH,W
	movlw	NUM_LEDS-1
	btfss	STATUS,C
	movf	ADRESH,W

	call	setwth
	goto adc_test

	return
